# Copyright 2024 Department of Computational Biology for Infection Research - Helmholtz Centre for Infection Research
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import collections
import os, sys, inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(currentdir)
sys.path.insert(0, parentdir)
import pandas as pd
import numpy as np


def transform_confusion_matrix2(query_w_length, confusion_df, precision_df, gs_df, log_scale=False):
    gs_df = gs_df.rename(columns={'BINID': 'genome_id', 'LENGTH': 'seq_length'})
    genomes_df = pd.DataFrame(gs_df['genome_id'].drop_duplicates(), columns=['genome_id'])

    all_genomes_sorted_by_size = gs_df.groupby('genome_id', sort=False).agg({'seq_length': 'sum'}).sort_values(by='seq_length', axis=0, ascending=False).index.tolist()

    genomes_sorted_list = precision_df[['genome_id', 'tp_length']].sort_values(by='tp_length', ascending=False).groupby(['genome_id'], sort=False).first().index.tolist()
    unmapped_genomes = set(genomes_df['genome_id']) - set(genomes_sorted_list)
    unmapped_genomes = [genome_id for genome_id in all_genomes_sorted_by_size if genome_id in unmapped_genomes]
    genomes_sorted_list += unmapped_genomes

    unbinned_genomes_df = genomes_df[~genomes_df['genome_id'].isin(confusion_df.reset_index()['genome_id'])]
    heatmap_df = pd.concat([confusion_df.reset_index(), unbinned_genomes_df], sort=True).fillna({'genome_length': 0, 'genome_seq_counts': 0})
    heatmap_df = heatmap_df.pivot(index='BINID', columns='genome_id', values='genome_length')
    heatmap_df = heatmap_df.merge(precision_df['tp_length'], on='BINID', sort=False).sort_values(by='tp_length', axis=0, ascending=False)

    gs_df = gs_df.set_index(['SEQUENCEID', 'genome_id'])
    query_df = query_w_length.set_index(['SEQUENCEID', 'genome_id'])
    difference_df = gs_df.index.difference(query_df.index)
    unassigned_sequences = gs_df.loc[difference_df].groupby('genome_id', sort=False).agg({'seq_length': 'sum'})

    heatmap_df = pd.concat([heatmap_df, unassigned_sequences.T], sort=True)
    heatmap_df = heatmap_df[genomes_sorted_list].fillna(0)
    if log_scale:
        heatmap_df = heatmap_df.apply(np.log10).replace(-np.inf, 0)

    return heatmap_df.astype(pd.SparseDtype(np.uint32, 0))


def transform_confusion_matrix(query):
    gold_standard = query.gold_standard

    bin_id_to_genome_id_to_length = {}
    for bin in query.bins:
        bin_id_to_genome_id_to_length[bin.id] = bin.mapping_id_to_length
    df_confusion = pd.DataFrame(bin_id_to_genome_id_to_length).T

    query_sequence_ids = set(query.sequence_id_to_bin_id.keys())
    gs_sequence_ids = set(gold_standard.sequence_id_to_bin_id.keys())
    genome_id_to_unassigned_bps = collections.Counter()
    for unassigned_seq_id in gs_sequence_ids - query_sequence_ids:
        genome_id = gold_standard.sequence_id_to_bin_id[unassigned_seq_id]
        genome_id_to_unassigned_bps[genome_id] += gold_standard.sequence_id_to_length[unassigned_seq_id]

    df_unassigned = pd.DataFrame.from_dict(genome_id_to_unassigned_bps, orient='index').rename(columns={0: 'unassigned'}).T
    table = pd.concat([df_confusion, df_unassigned], sort=False)
    table.fillna(0, inplace=True)
    # use log scale
    # table = table.applymap(np.log).fillna(0)

    bin_id_to_mapped_genome = {}
    for bin in query.bins:
        bin_id_to_mapped_genome[bin.id] = bin.most_abundant_genome

    # sort bins by the number of true positives (length of mapped genome within the bin)
    bin_id_to_mapped_genome_by_length = collections.OrderedDict(sorted(bin_id_to_mapped_genome.items(), key=lambda t: bin_id_to_genome_id_to_length[t[0]][t[1]], reverse=True))

    # sort genomes
    genome_order = []
    for bin_id in bin_id_to_mapped_genome_by_length:
        mapped_genome = bin_id_to_mapped_genome_by_length[bin_id]
        if mapped_genome not in genome_order:
            genome_order = pd.concat([genome_order, mapped_genome])
    genome_order += list(set(table.columns.values.tolist()) - set(genome_order))
    for genome_id in genome_id_to_unassigned_bps.keys():
        if genome_id not in genome_order:
            genome_order = pd.concat([genome_order, genome_id])

    table = table.loc[list(bin_id_to_mapped_genome_by_length.keys()) + ['unassigned'], genome_order]

    for genome_id in gold_standard.get_bin_ids():
        if genome_id not in table.columns.values.tolist():
            table[genome_id] = 0

    return table



# Copyright 2024 Department of Computational Biology for Infection Research - Helmholtz Centre for Infection Research
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import pandas as pd
import sys
import logging
import traceback
import os
import mimetypes
import gzip
import io
import tarfile
import zipfile
from multiprocessing.pool import ThreadPool
from collections import defaultdict
from collections import OrderedDict
from cami_amber import binning_classes
from cami_amber.utils import labels as utils_labels

try:
    import load_ncbi_taxinfo
except ImportError:
    sys.path.append(os.path.dirname(__file__))
    try:
        import load_ncbi_taxinfo
    finally:
        sys.path.remove(os.path.dirname(__file__))


def open_generic(file):
    file_type, file_encoding = mimetypes.guess_type(file)

    try:
        if file_encoding == 'gzip':
            if file_type == 'application/x-tar':  # .tar.gz
                tar = tarfile.open(file, 'r:gz')
                f = tar.extractfile(tar.getmembers()[0])
                return io.TextIOWrapper(f)
            else:  # .gz
                return gzip.open(file, 'rt')
        if file_type == 'application/zip':  # .zip
            f = zipfile.ZipFile(file, 'r')
            f = f.open(f.namelist()[0])
            return io.TextIOWrapper(f)
        else:
            return open(file, 'rt')
    except FileNotFoundError as e:
        logging.getLogger('amber').error('File not found: %s' % file)
        raise e


def open_coverages(file_path):
    if not file_path:
        return pd.DataFrame()
    logging.getLogger('amber').info('Loading coverage file')
    coverages_pd = pd.DataFrame()
    try:
        samples_metadata = read_metadata(file_path)
        for metadata in samples_metadata:
            nrows = metadata[1] - metadata[0] + 1
            df = pd.read_csv(file_path, sep='\t', comment='#', skiprows=metadata[0], nrows=nrows, header=None)
            df.rename(columns=metadata[3], inplace=True)
            df = df[['GENOMEID', 'COVERAGE']]
            df['SAMPLEID'] = metadata[2]['SAMPLEID']
            coverages_pd = pd.concat([coverages_pd, df], ignore_index=True)
    except BaseException as e:
        traceback.print_exc()
        logging.getLogger('amber').critical("File {} not found or malformed. {}".format(file_path, e))
        exit(1)
    return coverages_pd


def load_unique_common(unique_common_file_path, args_keyword):
    if not unique_common_file_path:
        return None
    logging.getLogger('amber').info('Loading list of genomes to be removed')
    with open(unique_common_file_path) as f:
        line = f.readline()
        line_split = line.split('\t')
        f.seek(0)
        if len(line_split) == 1:
            return [line.strip('\n') for line in f]
        elif len(line_split) > 1:
            genome_to_unique_common = {}
            for line in f:
                line_split = line.split('\t')
                genome_to_unique_common[line_split[0]] = line_split[1].strip('\n')
            if args_keyword:
                return [genome_id for genome_id, keyword in genome_to_unique_common.items() if keyword == args_keyword]
            else:
                return list(genome_to_unique_common.keys())
    return None


def load_ncbi_info(ncbi_dir):
    if ncbi_dir:
        logging.getLogger('amber').info('Loading NCBI taxonomy from %s' % ncbi_dir)
        try:
            if os.path.isfile(ncbi_dir):
                taxonomy_df = pd.read_feather(ncbi_dir).set_index('TAXID')
            else:
                taxonomy_df = pd.read_feather(os.path.join(ncbi_dir, 'nodes.amber.ft')).set_index('TAXID')
        except BaseException:
            traceback.print_exc()
            logging.getLogger('amber').info('Preprocessed NCBI taxonomy file not found. Creating file {}'.format(os.path.join(ncbi_dir, 'nodes.amber.ft')))
            taxonomy_df = load_ncbi_taxinfo.preprocess_ncbi_tax(ncbi_dir)
        taxonomy_df = taxonomy_df.astype(dtype={rank: pd.UInt32Dtype() for rank in load_ncbi_taxinfo.RANKS})
        return taxonomy_df


def read_metadata(path_label_tuple):
    file_path_query, label = path_label_tuple
    header = {}
    columns_list = []
    got_column_indices = False

    samples_metadata = []

    with open_generic(file_path_query) as f:
        for i, line in enumerate(f):
            line = line.strip()

            if len(line) == 0 or line.startswith("#"):
                if got_column_indices and not reading_data:
                    data_start = i + 1
                continue

            # parse header with column indices
            if line.startswith('@@'):
                for index, column_name in enumerate(line[2:].split('\t')):
                    columns_list.append(column_name)
                got_column_indices = True
                reading_data = False
                data_start = i + 1

            elif line.startswith('@'):
                logging.getLogger('amber').info('Found {} in {}'.format(line, label))
                # parse header with metadata
                if got_column_indices:
                    samples_metadata.append((data_start, data_end, header, columns_list, file_path_query, label))
                    header = {}
                    columns_list = []
                key, value = line[1:].split(':', 1)
                header[key.upper()] = value.strip()
                got_column_indices = False
                reading_data = False

            else:
                reading_data = True
                data_end = i
    try:
        samples_metadata.append((data_start, data_end, header, columns_list, file_path_query, label))
    except UnboundLocalError as e:
        logging.getLogger('amber').critical("File {} is malformed.".format(file_path_query))
        raise e
    return samples_metadata


def load_sample(metadata):
    columns = ['SEQUENCEID', 'BINID', 'TAXID', 'LENGTH', '_LENGTH']
    logging.getLogger('amber').info('Loading %s of %s' % (metadata[2]['SAMPLEID'], metadata[5]))
    nrows = metadata[1] - metadata[0] + 1
    usecols = [v for v in metadata[3] if v in columns]
    df = pd.read_csv(metadata[4], sep='\t', comment='#', skiprows=metadata[0], nrows=nrows, header=None,
                     names=metadata[3],
                     usecols=usecols,
                     dtype={'SEQUENCEID': pd.StringDtype(), 'BINID': pd.StringDtype(), 'TAXID': pd.UInt32Dtype(),
                            'LENGTH': pd.UInt32Dtype(), '_LENGTH': pd.UInt32Dtype()})
    df.rename(columns={'_LENGTH': 'LENGTH'}, inplace=True)
    return df


def load_binnings(samples_metadata, file_path_query):
    columns = ['SEQUENCEID', 'BINID', 'TAXID', 'LENGTH', '_LENGTH']
    sample_id_to_query_df = OrderedDict()
    for metadata in samples_metadata:
        logging.getLogger('amber').info('Loading ' + metadata[2]['SAMPLEID'])
        nrows = metadata[1] - metadata[0] + 1
        col_indices = [k for k, v in metadata[3].items() if v in columns]
        df = pd.read_csv(file_path_query, sep='\t', comment='#', skiprows=metadata[0], nrows=nrows, header=None, usecols=col_indices)
        df.rename(columns=metadata[3], inplace=True)
        df.rename(columns={'_LENGTH': 'LENGTH'}, inplace=True)
        sample_id_to_query_df[metadata[2]['SAMPLEID']] = df
    return sample_id_to_query_df


def open_query(file_path_query):
    try:
        samples_metadata = read_metadata(file_path_query)
        sample_id_to_query_df = load_binnings(samples_metadata, file_path_query)
    except BaseException as e:
        traceback.print_exc()
        logging.getLogger('amber').critical("File {} not found or malformed. {}".format(file_path_query, e))
        exit(1)
    return sample_id_to_query_df


def get_sample_id_to_num_genomes(sample_id_to_queries_list):
    sample_id_to_num_genomes = {}
    for sample_id in sample_id_to_queries_list:
        if isinstance(sample_id_to_queries_list[sample_id][0], binning_classes.GenomeQuery):
            gs_df = sample_id_to_queries_list[sample_id][0].df
            sample_id_to_num_genomes[sample_id] = gs_df['BINID'].nunique()
        else:
            sample_id_to_num_genomes[sample_id] = 0
    return sample_id_to_num_genomes


def get_rank_to_df(query_df, taxonomy_df, label, is_gs=False):
    if is_gs:
        if 'LENGTH' not in query_df.columns:
            logging.getLogger('amber').critical("Sequences length could not be determined. Please add column LENGTH to gold standard.")
            exit(1)
        cols = ['TAXID', 'LENGTH']
    else:
        cols = ['TAXID']

    rank_to_df = dict()
    logging.getLogger('amber').info('Setting taxa ranks')
    query_df = pd.merge(query_df, taxonomy_df.reset_index(), on=['TAXID']).set_index('SEQUENCEID')

    for rank in load_ncbi_taxinfo.RANKS:
        logging.getLogger('amber').info('Deriving bins of {} at rank: {}'.format(label, rank))
        if query_df[rank].isnull().all():
            continue
        rank_to_df[rank] = query_df[query_df[rank].notnull()][cols + [rank]]
        rank_to_df[rank]['TAXID'] = rank_to_df[rank][rank]
        rank_to_df[rank] = rank_to_df[rank].drop(columns=rank).reset_index()

    return rank_to_df


def load_queries(gold_standard_file, bin_files, labels, options=None, options_gs=None):
    max_workers = min(len(labels) + 1, os.cpu_count() or 1)
    pool = ThreadPool(max_workers)

    try:
        metadata_all = pool.map(read_metadata, zip([gold_standard_file] + bin_files, [utils_labels.GS] + labels))
        pool.close()
    except BaseException:
        logging.getLogger('amber').error('An error occurred. Exiting.')
        exit(1)
    samples_metadata_gs = metadata_all[0]
    samples_metadata_queries = metadata_all[1:]

    if not options:
        options = binning_classes.Options()
    if not options_gs:
        options_gs = binning_classes.Options()
    taxonomy_df = load_ncbi_info(options.ncbi_dir) if options.ncbi_dir else pd.DataFrame()

    sample_id_to_g_queries_list = defaultdict(list)
    sample_id_to_t_queries_list = defaultdict(list)
    sample_id_to_g_gs = defaultdict(list)
    sample_id_to_t_gs = defaultdict(list)

    for metadata in samples_metadata_gs:
        sample_id = metadata[2]['SAMPLEID']
        columns = metadata[3]
        if 'BINID' in columns:
            g_query_gs = binning_classes.GenomeQuery(utils_labels.GS, sample_id, options_gs, metadata, True)
            g_query_gs.gold_standard = g_query_gs
            sample_id_to_g_gs[sample_id] = g_query_gs
            if not options.skip_gs:
                sample_id_to_g_queries_list[sample_id].append(g_query_gs)
        if 'TAXID' in columns and not taxonomy_df.empty:
            t_query_gs = binning_classes.TaxonomicQuery(utils_labels.GS, sample_id, options_gs, metadata, taxonomy_df, True)
            t_query_gs.gold_standard = t_query_gs
            sample_id_to_t_gs[sample_id] = t_query_gs
            if not options.skip_gs:
                sample_id_to_t_queries_list[sample_id].append(t_query_gs)

    for query, label in zip(samples_metadata_queries, labels):
        for metadata in query:
            sample_id = metadata[2]['SAMPLEID']
            columns = metadata[3]
            if 'BINID' in columns:
                if not sample_id_to_g_gs[sample_id]:
                    logging.getLogger('amber').critical("Sample ID {} in {} not found in the genome binning gold standard.".format(sample_id, label))
                    exit(1)
                g_query = binning_classes.GenomeQuery(label, sample_id, options, metadata)
                g_query.gold_standard = sample_id_to_g_gs[sample_id]
                sample_id_to_g_queries_list[sample_id].append(g_query)
                options.only_taxonomic_queries = options_gs.only_taxonomic_queries = False
            if 'TAXID' in columns and not taxonomy_df.empty:
                if not sample_id_to_t_gs[sample_id]:
                    logging.getLogger('amber').critical("Sample ID {} in {} not found in the taxon binning gold standard.".format(sample_id, label))
                    exit(1)
                t_query = binning_classes.TaxonomicQuery(label, sample_id, options, metadata, taxonomy_df)
                t_query.gold_standard = sample_id_to_t_gs[sample_id]
                sample_id_to_t_queries_list[sample_id].append(t_query)
                options.only_genome_queries = options_gs.only_genome_queries = False

    return sample_id_to_g_queries_list, sample_id_to_t_queries_list, [metadata[2]['SAMPLEID'] for metadata in samples_metadata_gs]



# Copyright 2020 Department of Computational Biology for Infection Research - Helmholtz Centre for Infection Research
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from cami_amber.utils import labels as utils_labels
from cami_amber.utils import load_ncbi_taxinfo
from cami_amber import binning_classes
from cami_amber import plot_by_genome
import logging
import matplotlib
matplotlib.use('Agg')
import seaborn as sns
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
import matplotlib.ticker as ticker
import numpy as np
import os, sys, inspect
import pandas as pd
from collections import OrderedDict
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(currentdir)
sys.path.insert(0, parentdir)


def create_colors_list():
    colors_list = []
    for color in plt.cm.tab10(np.linspace(0, 1, 10))[:-1]:
        colors_list.append(tuple(color))
    colors_list.append("black")
    for color in plt.cm.Set2(np.linspace(0, 1, 8)):
        colors_list.append(tuple(color))
    for color in plt.cm.Set3(np.linspace(0, 1, 12)):
        colors_list.append(tuple(color))
    return colors_list


def create_legend(color_indices, available_tools, output_dir):
    colors_list = create_colors_list()
    if color_indices:
        colors_list = [colors_list[i] for i in color_indices]

    colors_iter = iter(colors_list)
    circles = [Line2D([], [], markeredgewidth=0.0, linestyle="None", marker="o", markersize=10, markerfacecolor=next(colors_iter)) for label in available_tools]

    fig = plt.figure(figsize=(0.5, 0.5))
    fig.legend(circles, available_tools, loc='center', frameon=False, ncol=5, handletextpad=0.1)
    fig.savefig(os.path.join(output_dir, 'genome', 'legend.pdf'), dpi=100, format='pdf', bbox_inches='tight')
    plt.close(fig)


def plot_precision_vs_bin_size(pd_bins, output_dir):
    pd_plot = pd_bins[pd_bins[utils_labels.TOOL] != utils_labels.GS]

    for tool_label, pd_tool in pd_plot.groupby(utils_labels.TOOL):
        fig, axs = plt.subplots(figsize=(5, 4.5))
        axs.scatter(np.log(pd_tool['total_length']), pd_tool['precision_bp'], marker='o')

        axs.set_xlim([None, np.log(pd_tool['total_length'].max())])
        axs.set_ylim([0.0, 1.0])
        axs.set_title(tool_label, fontsize=12)

        plt.ylabel('Purity per bin (%)', fontsize=12)
        plt.xlabel('Bin size [log(# bp)]', fontsize=12)

        fig.savefig(os.path.join(output_dir, 'genome', tool_label, 'purity_vs_bin_size.png'), dpi=200, format='png', bbox_inches='tight')
        plt.close(fig)


def plot_by_genome_coverage(pd_bins, pd_target_column, available_tools, output_dir):
    colors_list = create_colors_list()
    if len(available_tools) > len(colors_list):
        raise RuntimeError("Plot only supports 29 colors")

    fig, axs = plt.subplots(figsize=(5, 4.5))

    for i, (color, tool) in enumerate(zip(colors_list, available_tools)):
        pd_tool = pd_bins[pd_bins[utils_labels.TOOL] == tool].sort_values(by=['genome_index'])
        axs.scatter(pd_tool['genome_coverage'], pd_tool[pd_target_column], marker='o', color=colors_list[i], s=[3] * pd_tool.shape[0])
        window = 50
        rolling_mean = pd_tool[pd_target_column].rolling(window=window, min_periods=10).mean()
        axs.plot(pd_tool['genome_coverage'], rolling_mean, color=colors_list[i])

    axs.set_ylim([-0.01, 1.01])

    vals = axs.get_xticks()
    axs.xaxis.set_major_locator(ticker.FixedLocator(vals))
    axs.set_xticklabels(['{:,.1f}'.format(np.exp(x)) for x in vals], fontsize=12)

    vals = axs.get_yticks()
    axs.yaxis.set_major_locator(ticker.FixedLocator(vals))
    axs.set_yticklabels(['{:3.0f}'.format(x * 100) for x in vals], fontsize=12)

    axs.tick_params(axis='x', labelsize=12)

    if pd_target_column == 'precision_bp':
        ylabel = 'Purity per bin (%)'
        file_name = 'purity_by_genome_coverage'
    else:
        ylabel = 'Completeness per genome (%)'
        file_name = 'completeness_by_genome_coverage'

    plt.ylabel(ylabel, fontsize=15)
    plt.xlabel('Average genome coverage', fontsize=15)

    colors_iter = iter(colors_list)
    circles = []
    for x in range(len(available_tools)):
        circles.append(Line2D([], [], markeredgewidth=0.0, linestyle="None", marker="o", markersize=11, markerfacecolor=next(colors_iter)))
    lgd = plt.legend(circles, available_tools, bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0., handlelength=0, frameon=False, fontsize=14)

    fig.savefig(os.path.join(output_dir, 'genome', file_name + '.pdf'), dpi=100, format='pdf', bbox_extra_artists=(lgd,), bbox_inches='tight')
    plt.close(fig)


def get_pd_genomes_recall(sample_id_to_queries_list):
    pd_genomes_recall = pd.DataFrame()
    for sample_id in sample_id_to_queries_list:
        for query in sample_id_to_queries_list[sample_id]:
            if not isinstance(query, binning_classes.GenomeQuery):
                continue
            recall_df = query.recall_df_cami1[['genome_id', 'recall_bp']].copy()
            recall_df[utils_labels.TOOL] = query.label
            recall_df['sample_id'] = sample_id
            recall_df = recall_df.reset_index().set_index(['sample_id', utils_labels.TOOL])
            pd_genomes_recall = pd.concat([pd_genomes_recall, recall_df])
    return pd_genomes_recall


def plot_precision_recall_by_coverage(sample_id_to_queries_list, pd_bins_g, coverages_pd, available_tools, output_dir):
    # compute average genome coverage if coverages for multiple samples were provided
    coverages_pd = coverages_pd.groupby(['GENOMEID']).mean(numeric_only=True)
    coverages_pd.rename(columns={'GENOMEID': 'genome_id'})
    coverages_pd = coverages_pd.sort_values(by=['COVERAGE'])
    coverages_pd['rank'] = coverages_pd['COVERAGE'].rank()

    pd_genomes_recall = get_pd_genomes_recall(sample_id_to_queries_list)
    pd_genomes_recall['genome_index'] = pd_genomes_recall['genome_id'].map(coverages_pd['rank'].to_dict())
    pd_genomes_recall = pd_genomes_recall.reset_index()
    pd_genomes_recall['genome_coverage'] = np.log(pd_genomes_recall['genome_id'].map(coverages_pd['COVERAGE'].to_dict()))
    plot_by_genome_coverage(pd_genomes_recall, 'recall_bp', available_tools, output_dir)

    pd_bins_precision = pd_bins_g[[utils_labels.TOOL, 'precision_bp', 'genome_id']].copy().dropna(subset=['precision_bp'])
    pd_bins_precision['genome_index'] = pd_bins_precision['genome_id'].map(coverages_pd['rank'].to_dict())
    pd_bins_precision['genome_coverage'] = np.log(pd_bins_precision['genome_id'].map(coverages_pd['COVERAGE'].to_dict()))
    plot_by_genome_coverage(pd_bins_precision, 'precision_bp', available_tools, output_dir)


def plot_heatmap(df_confusion, sample_id, output_dir, label, separate_bar=False, log_scale=False):
    fig, axs = plt.subplots(figsize=(10, 8))
    fontsize = 20

    # replace columns and rows labels by numbers
    d = {value: key for (key, value) in enumerate(df_confusion.columns.tolist(), 1)}
    df_confusion = df_confusion.rename(index=str, columns=d)
    df_confusion.index = range(1, len(df_confusion) + 1)

    xticklabels = int(round(df_confusion.shape[1] / 10, -1))
    yticklabels = int(round(df_confusion.shape[0] / 10, -1))
    sns_plot = sns.heatmap(df_confusion, ax=axs, annot=False, cmap="YlGnBu_r", xticklabels=xticklabels, yticklabels=yticklabels, cbar=False, rasterized=True)

    # sns_plot = sns.heatmap(df_confusion, ax=axs, annot=False, cmap="YlGnBu_r", xticklabels=False, yticklabels=False, cbar=True, rasterized=True)
    sns_plot.set_xlabel("Genomes", fontsize=fontsize)
    sns_plot.set_ylabel("Predicted bins", fontsize=fontsize)
    plt.yticks(fontsize=12, rotation=0)
    plt.xticks(fontsize=12)

    mappable = sns_plot.get_children()[0]

    cbar_ax = fig.add_axes([.915, .11, .017, .77])
    if log_scale:
        fmt = lambda x, pos: '{:.0f}'.format((10 ** x) / 1000000)
    else:
        fmt = lambda x, pos: '{:.0f}'.format(x / 1000000)
    cbar = plt.colorbar(mappable, ax=axs, cax=cbar_ax, orientation='vertical', format=ticker.FuncFormatter(fmt))
    cbar.set_label(fontsize=fontsize, label='Millions of base pairs')

    cbar.ax.tick_params(labelsize=fontsize)
    cbar.outline.set_edgecolor(None)

    axs.set_title(label, fontsize=fontsize, pad=10)

    axs.set_ylim([len(df_confusion), 0])

    # plt.yticks(fontsize=14, rotation=0)
    # plt.xticks(fontsize=14)

    output_dir = os.path.join(output_dir, 'genome', label)

    fig.savefig(os.path.join(output_dir, 'heatmap_' + sample_id + '.pdf'), dpi=100, format='pdf', bbox_inches='tight')
    fig.savefig(os.path.join(output_dir, 'heatmap_' + sample_id + '.png'), dpi=200, format='png', bbox_inches='tight')
    plt.close(fig)

    if not separate_bar:
        return

    # create separate figure for bar
    fig = plt.figure(figsize=(6, 6))
    mappable = sns_plot.get_children()[0]
    fmt = lambda x, pos: '{:.0f}'.format(x / 1000000)

    cbar = plt.colorbar(mappable, orientation='vertical', label='[millions of base pairs]', format=ticker.FuncFormatter(fmt))

    text = cbar.ax.yaxis.label
    font = matplotlib.font_manager.FontProperties(size=16)
    text.set_font_properties(font)

    cbar.outline.set_visible(False)
    cbar.ax.tick_params(labelsize=14)

    # store separate bar figure
    plt.gca().set_visible(False)
    fig.savefig(os.path.join(output_dir, 'heatmap_bar.pdf'), dpi=100, format='pdf', bbox_inches='tight')

    plt.close(fig)


def plot_boxplot(sample_id_to_queries_list, metric_name, output_dir, available_tools):
    pd_bins = pd.DataFrame()
    for sample_id in sample_id_to_queries_list:
        for query in sample_id_to_queries_list[sample_id]:
            if not isinstance(query, binning_classes.GenomeQuery):
                continue
            metric_df = getattr(query, metric_name.replace('_bp', '_df')).copy()
            metric_df[utils_labels.TOOL] = query.label
            metric_df['sample_id'] = sample_id
            metric_df = metric_df.reset_index().set_index(['sample_id', utils_labels.TOOL])
            pd_bins = pd.concat([pd_bins, metric_df])

    metric_all = []

    for tool in available_tools:
        pd_tool = pd_bins.iloc[pd_bins.index.get_level_values(utils_labels.TOOL) == tool]
        metric_all.append(pd_tool[metric_name][pd_tool[metric_name].notnull()].tolist())

    fig, axs = plt.subplots(figsize=(6, 5))

    medianprops = dict(linewidth=2.5, color='gold')
    bplot = axs.boxplot(metric_all, notch=0, vert=0, patch_artist=True, labels=available_tools, medianprops=medianprops, sym='k.')
    colors_iter = iter(create_colors_list())

    # turn on grid
    axs.grid(which='major', linestyle=':', linewidth='0.5', color='lightgrey')

    # force axes to be from 0 to 100%
    axs.set_xlim([-0.01, 1.01])

    # transform plot_labels to percentages
    vals = axs.get_xticks()
    axs.xaxis.set_major_locator(ticker.FixedLocator(vals))
    axs.set_xticklabels(['{:3.0f}'.format(x * 100) for x in vals])

    # enable code to rotate labels
    tick_labels = axs.get_yticklabels()
    plt.setp(tick_labels, fontsize=13) ## rotation=55

    for box in bplot['boxes']:
        box.set(facecolor=next(colors_iter), linewidth=0.1)
    plt.ylim(plt.ylim()[::-1])

    if metric_name == 'precision_bp':
        axs.set_xlabel('Purity per bin (%)', fontsize=13)
        metric_name = 'purity_bp'
    else:
        axs.set_xlabel('Completeness per genome (%)', fontsize=13)
        metric_name = 'completeness_bp'

    fig.savefig(os.path.join(output_dir, 'genome', 'boxplot_' + metric_name + '.pdf'), dpi=100, format='pdf', bbox_inches='tight')
    fig.savefig(os.path.join(output_dir, 'genome', 'boxplot_' + metric_name + '.png'), dpi=200, format='png', bbox_inches='tight')

    # remove labels but keep grid
    # axs.get_yaxis().set_ticklabels([])
    # for tic in axs.yaxis.get_major_ticks():
    #     tic.tick1line.set_visible(False)
    #     tic.tick2line.set_visible(False)
    #     tic.label1.set_visible(False)
    #     tic.label2.set_visible(False)
    # fig.savefig(os.path.join(output_dir, 'genome', 'boxplot_' + metric_name + '_wo_legend.pdf'), dpi=100, format='pdf', bbox_inches='tight')
    plt.close(fig)


def plot_summary(color_indices, df_results, labels, output_dir, rank, plot_type, file_name, xlabel, ylabel):
    available_tools = df_results[utils_labels.TOOL].unique()
    tools = [tool for tool in labels if tool in available_tools]

    colors_list = create_colors_list()
    if color_indices:
        colors_list = [colors_list[i] for i in color_indices]
    df_mean = df_results.groupby(utils_labels.TOOL).mean(numeric_only=True).reindex(tools)

    binning_type = df_results[utils_labels.BINNING_TYPE].iloc[0]

    if len(df_mean) > len(colors_list):
        raise RuntimeError("Plot only supports 29 colors")

    fig, axs = plt.subplots(figsize=(5, 4.5))

    # force axes to be from 0 to 100%
    axs.set_xlim([0.0, 1.0])
    axs.set_ylim([0.0, 1.0])

    if plot_type == 'e':
        for i, (tool, df_row) in enumerate(df_mean.iterrows()):
            axs.errorbar(df_row['precision_avg_bp'], df_row['recall_avg_bp_cami1'], xerr=df_row['precision_avg_bp_var'], yerr=df_row['recall_avg_bp_var_cami1'],
                         fmt='o',
                         ecolor=colors_list[i],
                         mec=colors_list[i],
                         mfc=colors_list[i],
                         capsize=3,
                         markersize=8)
    if plot_type == 'f':
        for i, (tool, df_row) in enumerate(df_mean.iterrows()):
            axs.errorbar(df_row['precision_avg_seq'], df_row['recall_avg_seq_cami1'],
                         xerr=df_row['precision_avg_seq_sem'] if not np.isnan(df_row['precision_avg_seq_sem']) else 0,
                         yerr=df_row['recall_avg_seq_sem_cami1'] if not np.isnan(df_row['recall_avg_seq_sem_cami1']) else 0,
                         fmt='o',
                         ecolor=colors_list[i],
                         mec=colors_list[i],
                         mfc=colors_list[i],
                         capsize=3,
                         markersize=8)
    if plot_type == 'w':
        for i, (tool, df_row) in enumerate(df_mean.iterrows()):
            axs.plot(df_row['precision_weighted_bp'], df_row['recall_weighted_bp'], marker='o', color=colors_list[i], markersize=10)
    if plot_type == 'x':
        for i, (tool, df_row) in enumerate(df_mean.iterrows()):
            axs.plot(df_row['precision_weighted_seq'], df_row['recall_weighted_seq'], marker='o', color=colors_list[i], markersize=10)
    elif plot_type == 'p':
        for i, (tool, df_row) in enumerate(df_mean.iterrows()):
            axs.plot(df_row['adjusted_rand_index_bp'], df_row['percentage_of_assigned_bps'], marker='o', color=colors_list[i], markersize=10)

    # turn on grid
    # axs.minorticks_on()
    axs.grid(which='major', linestyle=':', linewidth='0.5')
    # axs.grid(which='minor', linestyle=':', linewidth='0.5')

    # transform plot_labels to percentages
    if plot_type != 'p':
        vals = axs.get_xticks()
        axs.xaxis.set_major_locator(ticker.FixedLocator(vals))
        axs.set_xticklabels(['{:3.0f}'.format(x * 100) for x in vals], fontsize=11)
    else:
        axs.tick_params(axis='x', labelsize=12)
    vals = axs.get_yticks()
    axs.yaxis.set_major_locator(ticker.FixedLocator(vals))
    axs.set_yticklabels(['{:3.0f}'.format(x * 100) for x in vals], fontsize=11)

    if rank:
        file_name = rank + '_' + file_name
        plt.title(rank)
        ylabel = ylabel.replace('genome', 'taxon')

    plt.xlabel(xlabel, fontsize=13)
    plt.ylabel(ylabel, fontsize=13)
    plt.tight_layout()
    fig.savefig(os.path.join(output_dir, binning_type, file_name + '.eps'), dpi=100, format='eps', bbox_inches='tight')

    colors_iter = iter(colors_list)
    circles = []
    for x in range(len(df_mean)):
        circles.append(Line2D([], [], markeredgewidth=0.0, linestyle="None", marker="o", markersize=11, markerfacecolor=next(colors_iter)))
    lgd = plt.legend(circles, tools, bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0., handlelength=0, frameon=False, fontsize=12)

    fig.savefig(os.path.join(output_dir, binning_type, file_name + '.pdf'), dpi=100, format='pdf', bbox_extra_artists=(lgd,), bbox_inches='tight')
    fig.savefig(os.path.join(output_dir, binning_type, file_name + '.png'), dpi=200, format='png', bbox_extra_artists=(lgd,), bbox_inches='tight')
    plt.close(fig)


def plot_avg_precision_recall(colors, df_results, labels, output_dir, rank=None):
    plot_summary(colors,
                 df_results,
                 labels,
                 output_dir,
                 rank,
                 'e',
                 'avg_purity_completeness_bp',
                 'Average purity per bin (%)',
                 'Average completeness per genome (%)')
    plot_summary(colors,
                 df_results,
                 labels,
                 output_dir,
                 rank,
                 'f',
                 'avg_purity_completeness_seq',
                 'Average purity per bin (%)',
                 'Average completeness per genome (%)')


def plot_precision_recall(colors, summary_per_query, labels, output_dir, rank=None):
    plot_summary(colors,
                 summary_per_query,
                 labels,
                 output_dir,
                 rank,
                 'w',
                 'purity_recall_bp',
                 'Purity for sample (%)',
                 'Completeness for sample (%)')
    plot_summary(colors,
                 summary_per_query,
                 labels,
                 output_dir,
                 rank,
                 'x',
                 'purity_completeness_seq',
                 'Purity for sample (%)',
                 'Completeness for sample (%)')


def plot_adjusted_rand_index_vs_assigned_bps(colors, summary_per_query, labels, output_dir, rank=None):
    plot_summary(colors,
                 summary_per_query,
                 labels,
                 output_dir,
                 rank,
                 'p',
                 'ari_vs_assigned_bps',
                 'Adjusted Rand index',
                 'Percentage of binned base pairs')


def plot_taxonomic_results(df_summary_t, metrics_list, errors_list, file_name, output_dir):
    colors_list = ["#006cba", "#008000", "#ba9e00", "red"]

    for tool, pd_results in df_summary_t.groupby(utils_labels.TOOL):
        dict_metric_list = []
        for metric in metrics_list:
            rank_to_metric = OrderedDict([(k, .0) for k in load_ncbi_taxinfo.RANKS])
            dict_metric_list.append(rank_to_metric)
        dict_error_list = []
        for error in errors_list:
            rank_to_metric_error = OrderedDict([(k, .0) for k in load_ncbi_taxinfo.RANKS])
            dict_error_list.append(rank_to_metric_error)

        for index, row in pd_results.iterrows():
            for rank_to_metric, metric in zip(dict_metric_list, metrics_list):
                rank_to_metric[row[utils_labels.RANK]] = .0 if np.isnan(row[metric]) else row[metric]
            for rank_to_metric_error, error in zip(dict_error_list, errors_list):
                rank_to_metric_error[row[utils_labels.RANK]] = .0 if np.isnan(row[error]) else row[error]

        fig, axs = plt.subplots(figsize=(6, 5))

        # force axes to be from 0 to 100%
        axs.set_xlim([0, 7])
        axs.set_ylim([0.0, 1.0])
        x_values = range(len(load_ncbi_taxinfo.RANKS))

        y_values_list = []
        for rank_to_metric, color in zip(dict_metric_list, colors_list):
            y_values = list(rank_to_metric.values())
            axs.plot(x_values, y_values, color=color)
            y_values_list.append(y_values)

        for rank_to_metric_error, y_values, color in zip(dict_error_list, y_values_list, colors_list):
            sem = list(rank_to_metric_error.values())
            plt.fill_between(x_values, np.subtract(y_values, sem).tolist(), np.add(y_values, sem).tolist(), color=color, alpha=0.5)

        plt.xticks(x_values, load_ncbi_taxinfo.RANKS, rotation='vertical')

        vals = axs.get_yticks()
        axs.yaxis.set_major_locator(ticker.FixedLocator(vals))
        axs.set_yticklabels(['{:3.0f}%'.format(x * 100) for x in vals])

        lgd = plt.legend([utils_labels.LABELS[x] for x in metrics_list], loc=1, borderaxespad=0., handlelength=2, frameon=False)

        plt.tight_layout()
        fig.savefig(os.path.join(output_dir, 'taxonomic', tool, file_name + '.png'), dpi=100, format='png', bbox_extra_artists=(lgd,), bbox_inches='tight')
        fig.savefig(os.path.join(output_dir, 'taxonomic', tool, file_name + '.pdf'), dpi=100, format='pdf', bbox_extra_artists=(lgd,), bbox_inches='tight')
        plt.close(fig)


def create_contamination_column(pd_tool_bins):
    pd_tool_bins['newcolumn'] = 1 - pd_tool_bins['precision_bp']


def create_completeness_minus_contamination_column(pd_tool_bins):
    pd_tool_bins['newcolumn'] = pd_tool_bins['recall_bp'] + pd_tool_bins['precision_bp'] - 1


def plot_contamination(pd_bins, binning_type, title, xlabel, ylabel, create_column_function, output_dir):
    if len(pd_bins) == 0:
        return

    pd_bins_copy = pd_bins[[utils_labels.TOOL, 'precision_bp', 'recall_bp']].copy().dropna(subset=['precision_bp'])
    create_column_function(pd_bins_copy)

    colors_list = create_colors_list()

    fig, axs = plt.subplots(figsize=(6, 5))

    tools = pd_bins_copy[utils_labels.TOOL].unique().tolist()

    for color, tool in zip(colors_list, tools):
        pd_tool_bins = pd_bins_copy[pd_bins_copy[utils_labels.TOOL] == tool]
        pd_tool_bins = pd_tool_bins.sort_values(by='newcolumn', ascending=False).reset_index()
        pd_tool_bins = pd_tool_bins.drop(['index'], axis=1)

        axs.plot(list(range(1, len(pd_tool_bins) + 1)), pd_tool_bins['newcolumn'], color=color)

    min_value = pd_bins_copy['newcolumn'].min()
    axs.set_ylim(min_value if min_value < 1.0 else .9, 1.0)
    axs.set_xlim(1, None)
    axs.grid(which='major', linestyle='-', linewidth='0.5', color='lightgrey')

    # transform plot_labels to percentages
    vals = axs.get_yticks()
    axs.yaxis.set_major_locator(ticker.FixedLocator(vals))
    axs.set_yticklabels(['{:3.0f}'.format(y * 100) for y in vals])

    plt.xlabel(xlabel, fontsize=14)
    plt.ylabel(ylabel + ' [%]', fontsize=14)

    lgd = plt.legend(tools, bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0., handlelength=1, frameon=False, fontsize=12)

    plt.tight_layout()

    file_name = title.lower().replace(' ', '_').replace('-', 'minus').replace('|', '')
    fig.savefig(os.path.join(output_dir, binning_type, file_name + '.png'), dpi=100, format='png', bbox_extra_artists=(lgd,), bbox_inches='tight')
    fig.savefig(os.path.join(output_dir, binning_type, file_name + '.pdf'), dpi=100, format='pdf', bbox_extra_artists=(lgd,), bbox_inches='tight')
    plt.close(fig)


def get_number_of_hq_bins(tools, pd_bins):
    pd_counts = pd.DataFrame()
    pd_bins_copy = pd_bins[[utils_labels.TOOL, 'precision_bp', 'recall_bp']].copy().dropna(subset=['precision_bp'])
    for tool in tools:
        pd_tool_bins = pd_bins_copy[pd_bins_copy[utils_labels.TOOL] == tool]
        x50 = pd_tool_bins[(pd_tool_bins['recall_bp'] > .5) & (pd_tool_bins['precision_bp'] > .9)].shape[0]
        x70 = pd_tool_bins[(pd_tool_bins['recall_bp'] > .7) & (pd_tool_bins['precision_bp'] > .9)].shape[0]
        x90 = pd_tool_bins[(pd_tool_bins['recall_bp'] > .9) & (pd_tool_bins['precision_bp'] > .9)].shape[0]
        pd_tool_counts = pd.DataFrame([[x90, x70, x50]], columns=['>90%', '>70%', '>50%'], index=[tool])
        pd_counts = pd.concat([pd_counts, pd_tool_counts])
    return pd_counts


def get_number_of_hq_bins_by_score(tools, pd_bins):
    pd_counts = pd.DataFrame()
    pd_bins_copy = pd_bins[[utils_labels.TOOL, 'precision_bp', 'recall_bp']].copy().dropna(subset=['precision_bp'])
    pd_bins_copy['newcolumn'] = pd_bins_copy['recall_bp'] + 5 * (pd_bins_copy['precision_bp'] - 1)
    for tool in tools:
        pd_tool_bins = pd_bins_copy[pd_bins_copy[utils_labels.TOOL] == tool]
        x50 = pd_tool_bins[pd_tool_bins['newcolumn'] > .5].shape[0]
        x70 = pd_tool_bins[pd_tool_bins['newcolumn'] > .7].shape[0]
        x90 = pd_tool_bins[pd_tool_bins['newcolumn'] > .9].shape[0]
        x50 -= x70
        x70 -= x90
        pd_tool_counts = pd.DataFrame([[x90, x70, x50]], columns=['>90', '>70', '>50'], index=[tool])
        pd_counts = pd.concat([pd_counts, pd_tool_counts])
    return pd_counts


def plot_counts(pd_bins, tools, output_dir, output_file, get_bin_counts_function):
    pd_counts = get_bin_counts_function(tools, pd_bins)
    fig, axs = plt.subplots(figsize=(11, 5))
    if output_file == 'bin_counts':
        fig = pd_counts.plot.bar(ax=axs, stacked=False, color=['#28334AFF', '#FBDE44FF', '#F65058FF'], width=.8, legend=None).get_figure()
    else:
        fig = pd_counts.plot.bar(ax=axs, stacked=True, color=['#9B4A97FF', '#FC766AFF', '#F9A12EFF'], width=.8, legend=None).get_figure()

    axs.tick_params(axis='x', labelrotation=45, length=0)
    axs.set_xticklabels(tools, horizontalalignment='right', fontsize=14)
    axs.set_xlabel(None)

    # axs.yaxis.set_major_locator(MaxNLocator(integer=True))

    h, l = axs.get_legend_handles_labels()
    axs.set_ylabel('#genome bins', fontsize=14)

    # axs.grid(which='major', linestyle=':', linewidth='0.5')
    # axs.grid(which='minor', linestyle=':', linewidth='0.5')

    ph = [plt.plot([], marker='', ls='')[0]]
    handles = ph + h

    if output_file == 'bin_counts':
        labels = ['Contamination < 10%           Completeness  '] + l
        bbox_to_anchor = (0.49, 1.02)
    else:
        labels = ['Score  '] + l
        y_values = (pd_counts['>90'] + pd_counts['>70'] + pd_counts['>50']).tolist()
        for i, v in enumerate(y_values):
            axs.text(i - .25, v + 5, str(v), color='black', fontweight='bold')
        bbox_to_anchor = (0.47, 1.02)

    lgd = plt.legend(handles, labels, bbox_to_anchor=bbox_to_anchor, columnspacing=.5, loc=8, borderaxespad=0., handlelength=1, frameon=False, fontsize=14, ncol=5)

    # plt.subplots_adjust(hspace=0.6, wspace=0.2)

    fig.savefig(os.path.join(output_dir, 'genome', output_file + '.pdf'), dpi=100, format='pdf', bbox_extra_artists=(lgd,), bbox_inches='tight')
    fig.savefig(os.path.join(output_dir, 'genome', output_file + '.png'), dpi=200, format='png', bbox_extra_artists=(lgd,), bbox_inches='tight')
    plt.close(fig)


def plot_genome_binning(color_indices, sample_id_to_queries_list, df_summary, pd_bins, labels, coverages_pd,
                        output_dir):
    df_summary_g = df_summary[df_summary[utils_labels.BINNING_TYPE] == 'genome']
    if len(df_summary_g) == 0:
        return

    logging.getLogger('amber').info('Creating genome binning plots')

    for sample_id in sample_id_to_queries_list:
        for query in sample_id_to_queries_list[sample_id]:
            query.plot()

    available_tools = list(df_summary_g[utils_labels.TOOL].unique())
    available_tools = [tool for tool in labels if tool in available_tools]

    if not coverages_pd.empty:
        plot_precision_recall_by_coverage(sample_id_to_queries_list, pd_bins, coverages_pd, available_tools,
                                                output_dir)

    if color_indices:
        color_indices = [int(i) - 1 for i in color_indices.split(',')]
    create_legend(color_indices, available_tools, output_dir)
    plot_avg_precision_recall(color_indices, df_summary_g, labels, output_dir)
    plot_precision_recall(color_indices, df_summary_g, labels, output_dir)
    plot_adjusted_rand_index_vs_assigned_bps(color_indices, df_summary_g, labels, output_dir)

    plot_boxplot(sample_id_to_queries_list, 'recall_bp', output_dir, available_tools)
    plot_boxplot(sample_id_to_queries_list, 'precision_bp', output_dir, available_tools)

    plot_counts(pd_bins, available_tools, output_dir, 'bin_counts', get_number_of_hq_bins)
    plot_counts(pd_bins, available_tools, output_dir, 'high_scoring_bins', get_number_of_hq_bins_by_score)

    plot_by_genome.plot_precision_recall_per_bin(pd_bins, output_dir)

    plot_contamination(pd_bins, 'genome', 'Contamination', 'Index of bin (sorted by contamination (bp))',
                       'Contamination (bp)', create_contamination_column, output_dir)
    plot_contamination(pd_bins, 'genome', 'Completeness - contamination',
                       'Index of bin (sorted by completeness - contamination (bp))',
                       'Completeness - contamination (bp)', create_completeness_minus_contamination_column,
                        output_dir)


def plot_taxonomic_binning(color_indices, df_summary, pd_bins, labels, output_dir):
    df_summary_t = df_summary[df_summary[utils_labels.BINNING_TYPE] == 'taxonomic']
    if len(df_summary_t) == 0:
        return

    logging.getLogger('amber').info('Creating taxonomic binning plots')

    available_tools = list(df_summary_t[utils_labels.TOOL].unique())
    available_tools = [tool for tool in labels if tool in available_tools]

    if color_indices:
        color_indices = [int(i) - 1 for i in color_indices.split(',')]
    for rank, pd_group in df_summary_t.groupby('rank'):
        plot_avg_precision_recall(color_indices, pd_group, available_tools, output_dir, rank)
        plot_precision_recall(color_indices, pd_group, available_tools, output_dir, rank)
        plot_adjusted_rand_index_vs_assigned_bps(color_indices, pd_group, available_tools, output_dir, rank)

    metrics_list = [utils_labels.AVG_PRECISION_BP, utils_labels.AVG_RECALL_BP]
    errors_list = [utils_labels.AVG_PRECISION_BP_SEM, utils_labels.AVG_RECALL_BP_SEM]
    plot_taxonomic_results(df_summary_t, metrics_list, errors_list, 'avg_precision_recall_bp', output_dir)

    metrics_list = [utils_labels.AVG_PRECISION_SEQ, utils_labels.AVG_RECALL_SEQ]
    errors_list = [utils_labels.AVG_PRECISION_SEQ_SEM, utils_labels.AVG_RECALL_SEQ_SEM]
    plot_taxonomic_results(df_summary_t, metrics_list, errors_list, 'avg_precision_recall_seq', output_dir)

    metrics_list = [utils_labels.PRECISION_PER_BP, utils_labels.RECALL_PER_BP, utils_labels.PRECISION_PER_SEQ,
                    utils_labels.RECALL_PER_SEQ]
    plot_taxonomic_results(df_summary_t, metrics_list, [], 'precision_recall', output_dir)

    for rank in load_ncbi_taxinfo.RANKS:
        pd_bins_rank = pd_bins[pd_bins['rank'] == rank]
        plot_contamination(pd_bins_rank, 'taxonomic', rank + ' | Contamination',
                           'Index of bin (sorted by contamination (bp))', 'Contamination (bp)',
                           create_contamination_column, output_dir)
        plot_contamination(pd_bins_rank, 'taxonomic', rank + ' | Completeness - contamination',
                           'Index of bin (sorted by completeness - contamination (bp))',
                           'Completeness - contamination (bp)',
                            create_completeness_minus_contamination_column, output_dir)



# Copyright 2023 Department of Computational Biology for Infection Research - Helmholtz Centre for Infection Research
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from collections import OrderedDict

TOOL = "Tool"
RANK = "rank"
BINNING_TYPE = "binning type"
SAMPLE = "Sample"
GS = 'Gold standard'
UNFILTERED = ' (unfiltered)'

AVG_PRECISION_BP = 'precision_avg_bp'
AVG_PRECISION_BP_SEM = 'precision_avg_bp_sem'

AVG_PRECISION_SEQ = 'precision_avg_seq'
AVG_PRECISION_SEQ_SEM = 'precision_avg_seq_sem'

AVG_RECALL_BP = 'recall_avg_bp'
AVG_RECALL_BP_CAMI1 = 'recall_avg_bp_cami1'
AVG_RECALL_BP_SEM = 'recall_avg_bp_sem'
AVG_RECALL_BP_SEM_CAMI1 = 'recall_avg_bp_sem_cami1'

AVG_RECALL_SEQ = 'recall_avg_seq'
AVG_RECALL_SEQ_CAMI1 = 'recall_avg_seq_cami1'
AVG_RECALL_SEQ_SEM = 'recall_avg_seq_sem'
AVG_RECALL_SEQ_SEM_CAMI1 = 'recall_avg_seq_sem_cami1'

F1_SCORE_BP = 'f1_score_bp'
F1_SCORE_SEQ = 'f1_score_seq'

F1_SCORE_BP_CAMI1 = 'f1_score_bp_cami1'
F1_SCORE_SEQ_CAMI1 = 'f1_score_seq_cami1'

PRECISION_PER_BP = 'precision_weighted_bp'
PRECISION_PER_SEQ = 'precision_weighted_seq'

RECALL_PER_BP = 'recall_weighted_bp'
RECALL_PER_SEQ = 'recall_weighted_seq'

F1_SCORE_PER_BP = 'f1_score_per_bp'
F1_SCORE_PER_SEQ = 'f1_score_per_seq'

RI_BY_BP = 'rand_index_bp'
RI_BY_SEQ = 'rand_index_seq'
ARI_BY_BP = 'adjusted_rand_index_bp'
ARI_BY_SEQ = 'adjusted_rand_index_seq'

PERCENTAGE_ASSIGNED_BPS = 'percentage_of_assigned_bps'
PERCENTAGE_ASSIGNED_SEQS = 'percentage_of_assigned_seqs'

ACCURACY_PER_BP = 'accuracy_bp'
ACCURACY_PER_SEQ = 'accuracy_seq'
MISCLASSIFICATION_PER_BP = 'misclassification_bp'
MISCLASSIFICATION_PER_SEQ = 'misclassification_seq'

UNIFRAC_BP = 'unifrac_bp'
UNIFRAC_SEQ = 'unifrac_seq'

QUALITY_OF_BINS = "Quality of bins: all bins have the same weight"
QUALITY_OF_SAMPLE = "Quality for sample"

TOOLTIP_AVG_PRECISION_BP = "Average fraction of base pairs in each predicted genome bin coming from the most abundant genome in each bin. It ranges from 0 (worst) to 1 (best)."
TOOLTIP_AVG_PRECISION_BP_TAX = "Average fraction of true positive base pairs in each predicted bin. It ranges from 0 (worst) to 1 (best)."
TOOLTIP_AVG_PRECISION_BP_SEM = "Standard error of the mean of average purity (bp)."

TOOLTIP_AVG_PRECISION_SEQ = "Average fraction of sequences in each predicted genome bin coming from the most abundant genome (in base pairs) in each bin. It ranges from 0 (worst) to 1 (best)."
TOOLTIP_AVG_PRECISION_SEQ_TAX = "Average fraction of true positive sequences in each predicted bin. It ranges from 0 (worst) to 1 (best)."
TOOLTIP_AVG_PRECISION_SEQ_SEM = "Standard error of the mean of average purity (seq)."

TOOLTIP_AVG_RECALL_BP = "Average fraction of base pairs of each genome, each of which have been put together in one predicted genome bin. For each genome, the bin that includes most of the base pairs from that genome is evaluated. This metric ranges from 0 (worst) to 1 (best)."
TOOLTIP_AVG_RECALL_BP_TAX = "Average fraction of true positive base pairs in each predicted bin from the true bin size. It ranges from 0 (worst) to 1 (best)."

TOOLTIP_AVG_RECALL_BP_CAMI1 = "Average completeness per genome in base pairs, as computed in AMBER v1 (Meyer et al. GigaScience 2018) and the 1st CAMI Challenge (Sczyrba et al. Nature Methods 2017). It ranges from 0 (worst) to 1 (best)."
TOOLTIP_AVG_RECALL_BP_SEM = "Standard error of the mean of average completeness (bp)."
TOOLTIP_AVG_RECALL_SEQ = "Average fraction of sequences of each genome, each of which have been put together in one predicted genome bin. For each genome, the bin that includes most of the base pairs from that genome is evaluated. This metric ranges from 0 (worst) to 1 (best)."
TOOLTIP_AVG_RECALL_SEQ_TAX = "Average fraction of true positive sequences in each predicted bin from the true bin size in number of sequences. It ranges from 0 (worst) to 1 (best)."

TOOLTIP_AVG_RECALL_SEQ_CAMI1 = "Average completeness per genome in number of sequences, as computed in AMBER v1 (Meyer et al. GigaScience 2018) and the 1st CAMI Challenge (Sczyrba et al. Nature Methods 2017). It ranges from 0 (worst) to 1 (best)."
TOOLTIP_AVG_RECALL_SEQ_SEM = "Standard error of the mean of average completeness (seq)."

TOOLTIP_PRECISION_PER_BP = "Sum of base pairs coming from the most abundant genome in each predicted genome bin divided by the sum of base pairs in all predicted bins. It ranges from 0 (worst) to 1 (best)."
TOOLTIP_PRECISION_PER_BP_TAX = "Sum of true positive base pairs in the predicted bins divided by the sum of base pairs in all predicted bins. It ranges from 0 (worst) to 1 (best)."

TOOLTIP_PRECISION_PER_SEQ = "Sum of #sequences coming from the most abundant genome (in base pairs) in each predicted genome bin divided by the sum of #sequences in all predicted bins. It ranges from 0 (worst) to 1 (best)."
TOOLTIP_PRECISION_PER_SEQ_TAX = "Sum of true positive sequences in the predicted bins divided by the sum of #sequences in all predicted bins. It ranges from 0 (worst) to 1 (best)."

TOOLTIP_RECALL_PER_BP = "Sum of base pairs of each genome, each of which have been put together in one predicted genome bin, divided by the sum of base pairs of all underlying genomes. For each genome, the bin that includes most of the base pairs from that genome is considered. It ranges from 0 (worst) to 1 (best)."
TOOLTIP_RECALL_PER_BP_TAX = "Sum of true positive base pairs in the predicted bins divided by the sum of base pairs in all underlying taxa. It ranges from 0 (worst) to 1 (best)."

TOOLTIP_RECALL_PER_SEQ = "Sum of #sequences of each genome, each of which have been put together in one predicted genome bin, divided by the sum of #sequences of all underlying genomes. For each genome, the bin that includes most of the base pairs from that genome is considered. It ranges from 0 (worst) to 1 (best)."
TOOLTIP_RECALL_PER_SEQ_TAX = "Sum of true positive sequences in the predicted bins divided by the sum of #sequences in all underlying taxa. It ranges from 0 (worst) to 1 (best)."

TOOLTIP_F1_SCORE_BP = "Two times the product of average purity (bp) and average completeness (bp) divided by their sum. It ranges from 0 (worst) to 1 (best)."
TOOLTIP_F1_SCORE_SEQ = "Two times the product of average purity (seq) and average completeness (seq) divided by their sum. It ranges from 0 (worst) to 1 (best)."
TOOLTIP_F1_SCORE_PER_BP = "Two times the product of purity (bp) and completeness (bp) divided by their sum. It ranges from 0 (worst) to 1 (best)."
TOOLTIP_F1_SCORE_PER_SEQ = "Two times the product of purity (seq) and completeness (seq) divided by their sum. It ranges from 0 (worst) to 1 (best)."

TOOLTIP_RI_BY_BP = "Overall resolution in base pairs of the underlying ground truth genomes on the binned part of the sample. It ranges from 0 (worst) to 1 (best)."
TOOLTIP_RI_BY_BP_TAX = "Overall resolution in base pairs of the underlying ground truth taxa on the binned part of the sample. It ranges from 0 (worst) to 1 (best)."

TOOLTIP_RI_BY_SEQ = "Overall resolution in sequences of the underlying ground truth genomes on the binned part of the sample. It ranges from 0 (worst) to 1 (best)."
TOOLTIP_RI_BY_SEQ_TAX = "Overall resolution in sequences of the underlying ground truth taxa on the binned part of the sample. It ranges from 0 (worst) to 1 (best)."

TOOLTIP_ARI_BY_BP = "Rand index (bp) ajusted by the expected Rand index of a random clustering. It ranges from 0 (worst) to 1 (best)."
TOOLTIP_ARI_BY_SEQ = "Rand index (seq) ajusted by the expected Rand index of a random clustering. It ranges from 0 (worst) to 1 (best)."
TOOLTIP_PERCENTAGE_ASSIGNED_BPS = "Fraction of base pairs from the complete sample that have been assigned to predicted bins. It ranges from 0 (worst) to 1 (best)."
TOOLTIP_PERCENTAGE_ASSIGNED_SEQS = "Fraction of sequences from the complete sample that have been assigned to predicted bins. It ranges from 0 (worst) to 1 (best)."

TOOLTIP_ACCURACY_PER_BP = "Sum of base pairs coming from the most abundant genome in each predicted genome bin divided by the sum of base pairs in the complete sample. It ranges from 0 (worst) to 1 (best)."
TOOLTIP_ACCURACY_PER_BP_TAX = "Equivalent to completeness (bp), i.e. sum of true positive base pairs in the predicted bins divided by the sum of base pairs in all underlying taxa. It ranges from 0 (worst) to 1 (best)."

TOOLTIP_ACCURACY_PER_SEQ = "Sum of #sequences coming from the most abundant genome in each predicted genome bin divided by the sum of #sequences in the complete sample. It ranges from 0 (worst) to 1 (best)."
TOOLTIP_ACCURACY_PER_SEQ_TAX = "Equivalent to completeness (seq), i.e. sum of true positive sequences in the predicted bins divided by the sum of #sequences in all underlying taxa. It ranges from 0 (worst) to 1 (best)."

TOOLTIP_MISCLASSIFICATION_PER_BP = "1 - purity (bp)"
TOOLTIP_MISCLASSIFICATION_PER_SEQ = "1 - purity (seq)"

TOOLTIP_UNIFRAC_BP = "Tree-based measure of similarity between the true and predicted base pair assignments at all taxonomic ranks ranging from 0 (high similarity) to 16 (low similarity)."
TOOLTIP_UNIFRAC_SEQ = "Tree-based measure of similarity between the true and predicted base pair assignments at all taxonomic ranks ranging from 0 (high similarity) to 16 (low similarity)."

LABELS = {'precision_avg_bp': 'Average purity (bp)',
          'precision_avg_seq': 'Average purity (seq)',
          'precision_avg_bp_sem': 'Std error of av. purity (bp)',
          'precision_avg_seq_sem': 'Std error of av. purity (seq)',
          'recall_avg_bp': 'Average completeness (bp)',
          'recall_avg_seq': 'Average completeness (seq)',
          'recall_avg_bp_cami1': 'CAMI 1 average completeness (bp)',
          'recall_avg_seq_cami1': 'CAMI 1 average completeness (seq)',
          'recall_avg_bp_sem': 'Std error of av. completeness (bp)',
          'recall_avg_seq_sem': 'Std error of av. completeness (seq)',
          'recall_avg_bp_sem_cami1': 'CAMI 1 std error of av. completeness (bp)',
          'recall_avg_seq_sem_cami1': 'CAMI 1 std error of av. completeness (seq)',
          'f1_score_bp': 'F1 score (bp)',
          'f1_score_seq': 'F1 score (seq)',
          'f1_score_bp_cami1': 'CAMI 1 F1 score (bp)',
          'f1_score_seq_cami1': 'CAMI 1 F1 score (seq)',
          'f1_score_per_bp': 'F1 score for sample (bp)',
          'f1_score_per_seq': 'F1 score for sample (seq)',
          'accuracy_bp': 'Accuracy (bp)',
          'accuracy_seq': 'Accuracy (seq)',
          'misclassification_bp': 'Misclassification rate (bp)',
          'misclassification_seq': 'Misclassification rate (seq)',
          'precision_weighted_bp': 'Purity (bp)',
          'precision_weighted_seq': 'Purity (seq)',
          'recall_weighted_bp': 'Completeness (bp)',
          'recall_weighted_seq': 'Completeness (seq)',
          'rand_index_bp': 'Rand index (bp)',
          'rand_index_seq': 'Rand index (seq)',
          'adjusted_rand_index_bp': 'Adjusted Rand index (bp)',
          'adjusted_rand_index_seq': 'Adjusted Rand index (seq)',
          'percentage_of_assigned_bps': 'Percentage of binned bp',
          'percentage_of_assigned_seqs': 'Percentage of binned sequences',
          'unifrac_bp': 'UniFrac (bp)',
          'unifrac_seq': 'UniFrac (seq)'}


LABELS1 = [TOOL,
           ACCURACY_PER_BP,
           ACCURACY_PER_SEQ,
           ARI_BY_BP,
           ARI_BY_SEQ,
           AVG_RECALL_BP,
           AVG_RECALL_SEQ,
           AVG_PRECISION_BP,
           AVG_PRECISION_SEQ,
           F1_SCORE_BP_CAMI1,
           F1_SCORE_SEQ_CAMI1,
           AVG_RECALL_BP_CAMI1,
           AVG_RECALL_SEQ_CAMI1,
           AVG_RECALL_BP_SEM_CAMI1,
           AVG_RECALL_SEQ_SEM_CAMI1,
           RECALL_PER_BP,
           RECALL_PER_SEQ,
           F1_SCORE_BP,
           F1_SCORE_SEQ,
           F1_SCORE_PER_BP,
           F1_SCORE_PER_SEQ,
           MISCLASSIFICATION_PER_BP,
           MISCLASSIFICATION_PER_SEQ,
           PERCENTAGE_ASSIGNED_BPS,
           PERCENTAGE_ASSIGNED_SEQS,
           PRECISION_PER_BP,
           PRECISION_PER_SEQ,
           RI_BY_BP,
           RI_BY_SEQ,
           SAMPLE,
           AVG_RECALL_BP_SEM,
           AVG_RECALL_SEQ_SEM,
           AVG_PRECISION_BP_SEM,
           AVG_PRECISION_SEQ_SEM,
           UNIFRAC_BP,
           UNIFRAC_SEQ,
           BINNING_TYPE,
           RANK]


def get_genome_bins_columns():
    return OrderedDict([('BINID', 'Bin ID'),
                        ('genome_id', 'Most abundant genome'),
                        ('precision_bp', LABELS[PRECISION_PER_BP]),
                        ('recall_bp', LABELS[RECALL_PER_BP]),
                        ('total_length', 'Bin size (bp)'),
                        ('tp_length', 'True positives (bp)'),
                        ('length_gs', 'True size of most abundant genome (bp)'),
                        ('precision_seq', LABELS[PRECISION_PER_SEQ]),
                        ('recall_seq', LABELS[RECALL_PER_SEQ]),
                        ('total_seq_counts', 'Bin size (seq)'),
                        ('tp_seq_counts', 'True positives (seq)'),
                        ('seq_counts_gs', 'True size of most abundant genome (seq)')])


def get_tax_bins_columns():
    return OrderedDict([('TAXID', 'Taxon ID'),
                        ('name', 'Scientific name'),
                        ('rank', 'Taxonomic rank'),
                        ('precision_bp', LABELS[PRECISION_PER_BP]),
                        ('recall_bp', LABELS[RECALL_PER_BP]),
                        ('total_length', 'Bin size (bp)'),
                        ('tp_length', 'True positives (bp)'),
                        ('length_gs', 'True size (bp)'),
                        ('precision_seq', LABELS[PRECISION_PER_SEQ]),
                        ('recall_seq', LABELS[RECALL_PER_SEQ]),
                        ('total_seq_counts', 'Bin size (seq)'),
                        ('tp_seq_counts', 'True positives (seq)'),
                        ('seq_counts_gs', 'True size (seq)'),
                        ('filtered', 'Filtered')])



#!/usr/bin/env python3

# Copyright 2024 Department of Computational Biology for Infection Research - Helmholtz Centre for Infection Research
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from cami_amber import amber_html
from cami_amber import evaluate
from cami_amber import plots
from cami_amber import binning_classes
from cami_amber.utils import load_data
from cami_amber.utils import argparse_parents
from cami_amber.utils import labels as utils_labels
from version import __version__
from collections import defaultdict
import argparse
import errno
import logging
import os
import sys
import pandas as pd


def get_logger(output_dir, silent):
    make_sure_path_exists(output_dir)
    logger = logging.getLogger('amber')
    logger.setLevel(logging.INFO)
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    logging_fh = logging.FileHandler(os.path.join(output_dir, 'log.txt'))
    logging_fh.setFormatter(formatter)
    logger.addHandler(logging_fh)

    if not silent:
        logging_stdout = logging.StreamHandler(sys.stdout)
        logging_stdout.setFormatter(formatter)
        logger.addHandler(logging_stdout)
    return logger


def make_sure_path_exists(path):
    try:
        os.makedirs(path)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise


def create_output_directories(output_dir, sample_id_to_queries_list):
    logging.getLogger('amber').info('Creating output directories')
    for sample_id in sample_id_to_queries_list:
        for query in sample_id_to_queries_list[sample_id]:
            make_sure_path_exists(os.path.join(output_dir, query.binning_type, query.label))


def get_labels(labels, bin_files):
    if labels:
        labels_list = [x.strip() for x in labels.split(',')]
        if len(set(labels_list)) != len(bin_files):
            logging.getLogger('amber').critical(
                'Number of different labels does not match the number of binning files. Please check parameter -l, --labels.')
            exit(1)
        return labels_list
    tool_id = []
    for bin_file in bin_files:
        tool_id.append(bin_file.split('/')[-1].split('.binning')[0])
    return tool_id


def save_metrics(sample_id_to_queries_list, df_summary, pd_bins, output_dir, stdout):
    logging.getLogger('amber').info('Saving computed metrics')
    df_summary.to_csv(os.path.join(output_dir, 'results.tsv'), sep='\t', index=False)
    pd_bins.to_csv(os.path.join(output_dir, 'bin_metrics.tsv'), index=False, sep='\t')
    if stdout:
        print(df_summary[[label for label in utils_labels.LABELS1 if label in df_summary.columns]].rename(columns=utils_labels.LABELS).to_string(index=False))
    for tool, pd_group in pd_bins[pd_bins['rank'] == 'NA'].groupby(utils_labels.TOOL):
        bins_columns = utils_labels.get_genome_bins_columns()
        table = pd_group[['sample_id'] + list(bins_columns.keys())].rename(columns=dict(bins_columns))
        table.to_csv(os.path.join(output_dir, 'genome', tool, 'metrics_per_bin.tsv'), sep='\t', index=False)
    for tool, pd_group in pd_bins[pd_bins['rank'] != 'NA'].groupby(utils_labels.TOOL):
        bins_columns = utils_labels.get_tax_bins_columns()
        if 'name' not in pd_bins.columns or pd_group['name'].isnull().any():
            del bins_columns['name']
        table = pd_group[['sample_id'] + list(bins_columns.keys())].rename(columns=dict(bins_columns))
        table.to_csv(os.path.join(output_dir, 'taxonomic', tool, 'metrics_per_bin.tsv'), sep='\t', index=False)

    pd_genomes_all = pd.DataFrame()
    for sample_id in sample_id_to_queries_list:
        pd_genomes_sample = pd.DataFrame()
        for query in sample_id_to_queries_list[sample_id]:
            if isinstance(query, binning_classes.GenomeQuery):
                query.recall_df_cami1[utils_labels.TOOL] = query.label
                pd_genomes_sample = pd.concat([pd_genomes_sample, query.recall_df_cami1], ignore_index=True, sort=False)
        pd_genomes_sample['sample_id'] = sample_id
        pd_genomes_all = pd.concat([pd_genomes_all, pd_genomes_sample], ignore_index=True, sort=False)
    if not pd_genomes_all.empty:
        pd_genomes_all.to_csv(os.path.join(output_dir, 'genome_metrics_cami1.tsv'), index=False, sep='\t')


def main(args=None):
    parser = argparse.ArgumentParser(description="AMBER: Assessment of Metagenome BinnERs",
                                     parents=[argparse_parents.PARSER_MULTI2], prog='AMBER')
    parser.add_argument('-p', '--filter', help=argparse_parents.HELP_FILTER)
    parser.add_argument('-n', '--min_length', help="Minimum length of sequences", type=int, required=False)
    parser.add_argument('-o', '--output_dir', help="Directory to write the results to", required=True)
    parser.add_argument('--stdout', help="Print summary to stdout", action='store_true')
    parser.add_argument('-d', '--desc', help="Description for HTML page", required=False)
    parser.add_argument('--colors', help="Color indices", required=False)
    parser.add_argument('--silent', help='Silent mode', action='store_true')
    parser.add_argument('--skip_gs', help='Skip gold standard evaluation vs itself', action='store_true')
    parser.add_argument('-v', '--version', action='version', version='%(prog)s ' + __version__)

    group_g = parser.add_argument_group('genome binning-specific arguments')
    group_g.add_argument('-x', '--min_completeness', help=argparse_parents.HELP_THRESHOLDS_COMPLETENESS, required=False)
    group_g.add_argument('-y', '--max_contamination', help=argparse_parents.HELP_THRESHOLDS_CONTAMINATION,
                         required=False)
    group_g.add_argument('-r', '--remove_genomes', help=argparse_parents.HELP_GENOMES_FILE, required=False)
    group_g.add_argument('-k', '--keyword', help=argparse_parents.HELP_KEYWORD, required=False)
    group_g.add_argument('--genome_coverage', help='genome coverages', required=False)

    group_t = parser.add_argument_group('taxonomic binning-specific arguments')
    group_t.add_argument('--ncbi_dir', help="Directory containing the NCBI taxonomy database dump files nodes.dmp, merged.dmp, and names.dmp", required=False)
    # group_t.add_argument('--rank_as_genome_binning',
    #                      help="Assess taxonomic binning at a rank also as genome binning. Valid ranks: superkingdom, phylum, class, order, family, genus, species, strain",
    #                      required=False)

    args = parser.parse_args(args)
    output_dir = os.path.abspath(args.output_dir)
    logger = get_logger(output_dir, args.silent)

    labels = get_labels(args.labels, args.bin_files)

    genome_to_unique_common = load_data.load_unique_common(args.remove_genomes, args.keyword)

    options = binning_classes.Options(filter_tail_percentage=args.filter,
                                      genome_to_unique_common=genome_to_unique_common,
                                      filter_keyword=args.keyword,
                                      min_length=args.min_length,
                                      rank_as_genome_binning=None, #args.rank_as_genome_binning,
                                      output_dir=output_dir,
                                      min_completeness=args.min_completeness,
                                      max_contamination=args.max_contamination,
                                      ncbi_dir=args.ncbi_dir,
                                      skip_gs=args.skip_gs)
    options_gs = binning_classes.Options(filter_tail_percentage=.0,
                                         genome_to_unique_common=genome_to_unique_common,
                                         filter_keyword=args.keyword,
                                         min_length=args.min_length,
                                         rank_as_genome_binning=None, #args.rank_as_genome_binning,
                                         output_dir=output_dir,
                                         ncbi_dir=args.ncbi_dir,
                                         skip_gs=args.skip_gs)

    sample_id_to_g_queries_list, sample_id_to_t_queries_list, sample_ids_list = load_data.load_queries(
        args.gold_standard_file, args.bin_files, labels, options, options_gs)

    coverages_pd = load_data.open_coverages(args.genome_coverage)

    sample_id_to_queries_list = defaultdict(list)
    for sample_id in sample_id_to_g_queries_list:
        sample_id_to_queries_list[sample_id] += sample_id_to_g_queries_list[sample_id]
    for sample_id in sample_id_to_t_queries_list:
        sample_id_to_queries_list[sample_id] += sample_id_to_t_queries_list[sample_id]

    create_output_directories(output_dir, sample_id_to_queries_list)

    df_summary, pd_bins = evaluate.evaluate_samples_queries(sample_id_to_g_queries_list, sample_id_to_t_queries_list)

    save_metrics(sample_id_to_queries_list, df_summary, pd_bins, output_dir, args.stdout)

    plots.plot_genome_binning(args.colors,
                              sample_id_to_g_queries_list,
                              df_summary,
                              pd_bins[pd_bins['rank'] == 'NA'],
                              labels,
                              coverages_pd,
                              output_dir)
    plots.plot_taxonomic_binning(args.colors, df_summary, pd_bins, labels, output_dir)

    amber_html.create_html(df_summary,
                           pd_bins,
                           [utils_labels.GS] + labels,
                           sample_ids_list,
                           options,
                           args.desc)
    logger.info('AMBER finished successfully. All results have been saved to {}'.format(output_dir))


if __name__ == "__main__":
    main()



import mimetypes
import gzip
import io
import tarfile
import zipfile


print(mimetypes.guess_type('/home/fmeyer/tmp/test/test.fasta'))
print('.zip\t', mimetypes.guess_type('/home/fmeyer/tmp/test.zip'))
print('.gz\t', mimetypes.guess_type('/home/fmeyer/tmp/test.gz')) # gzip.open
print('tar.gz\t', mimetypes.guess_type('/home/fmeyer/tmp/test.tar.gz'))


def open_generic(file):
    file_type, file_encoding = mimetypes.guess_type(file)

    if file_encoding == 'gzip':
        if file_type == 'application/x-tar':  # .tar.gz
            tar = tarfile.open(file, 'r:gz')
            f = tar.extractfile(tar.getmembers()[0])
            return io.TextIOWrapper(f)
        else:  # .gz
            return gzip.open(file, 'rt')
    if file_type == 'application/zip':  # .zip
        f = zipfile.ZipFile(file, 'r')
        f = f.open(f.namelist()[0])
        return io.TextIOWrapper(f)
    else:
        return open(file, 'rt')


ff = None
with open_generic('/home/fmeyer/tmp/test/test.fasta.gz') as ff:
    for line in ff:
        print(line, end='')
    #     break
    # print(ff.closed)

print(ff.closed)



# This is an example of how to use the script
from cami_amber.utils import ProfilingTools as PF
from cami_amber.utils import EMDUnifrac as EMDU
import numpy as np
import argparse
import copy
import sys


def compute_unifrac(pf1, pf2):
    P1 = copy.deepcopy(pf1)
    P2 = copy.deepcopy(pf2)
    (Tint, lint, nodes_in_order, nodes_to_index, P, Q) = P1.make_unifrac_input_and_normalize(P2)
    (weighted, _) = EMDU.EMDUnifrac_weighted(Tint, lint, nodes_in_order, P, Q)
    # (unweighted, _) = EMDU.EMDUnifrac_unweighted(Tint, lint, nodes_in_order, P, Q)
    return weighted #, unweighted


def weighted_unifrac(pf1, pf2):
    P1 = copy.deepcopy(pf1)
    P2 = copy.deepcopy(pf2)
    (Tint, lint, nodes_in_order, nodes_to_index, P, Q) = P1.make_unifrac_input_and_normalize(P2)
    (val, _) = EMDU.EMDUnifrac_weighted(Tint, lint, nodes_in_order, P, Q)
    return val


def unweighted_unifrac(pf1, pf2):
    P1 = copy.deepcopy(pf1)
    P2 = copy.deepcopy(pf2)
    (Tint, lint, nodes_in_order, nodes_to_index, P, Q) = P1.make_unifrac_input_and_normalize(P2)
    (val, _) = EMDU.EMDUnifrac_unweighted(Tint, lint, nodes_in_order, P, Q)
    return val


def print_list_unifrac(unifrac_list, labels, stream=sys.stdout):
    labels_iterator = iter(labels)
    stream.write("tool\tweighted\tunweighted\n")
    for unifrac in unifrac_list:
        stream.write("{}\t".format(next(labels_iterator)))
        stream.write("\t".join([format(unifrac[0], '.3f'), format(unifrac[1], '.3f')]))
        stream.write("\n")


def read_params(args):
	parser = argparse.ArgumentParser(description='')
	arg = parser.add_argument
	arg('--input', metavar='files_file', type=str, required=True,
					default=None, help="File of CAMI profile files to use")
	arg('--output', metavar='output_file', required=True, default=None, type=str,
					help="Output file (you should have this end in .csv as it is a matrix)")
	arg('--threshold', metavar='threshold', required=True, default=None, type=float,
					help="Value to threshold profiles to before computing EMDUnifrac. "
					"NOTE THIS VALUE IS IN PERCENTAGES so if you want 1% use 1")
	return vars(parser.parse_args())


if __name__ == '__main__':
	par = read_params(sys.argv)
	files_file = par['input']
	output_file = par['output']
	threshold = par['threshold']

	# Get all the profile file names
	files = []
	fid = open(files_file, 'r')
	for line in fid.readlines():
		files.append(line.strip())
	fid.close()

	# Import all the profiles
	profiles = []
	for file_name in files:
		profiles.append(PF.Profile(input_file_name=file_name))

	# Threshold all the profiles
	for profile in profiles:
		profile.threshold(threshold=threshold)

	# Compute EMDUnifrac
	D = np.zeros((len(profiles), len(profiles)))
	for i in range(len(profiles)):
		for j in range(i+1, len(profiles)):
			(wu, uu) = compute_unifrac(profiles[i], profiles[j])
			D[i, j] = wu
			D[j, i] = uu

	# Save results in tsv
	np.savetxt(output_file, D, delimiter='\t', newline='\n')



# Copyright 2023 Department of Computational Biology for Infection Research - Helmholtz Centre for Infection Research
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import matplotlib
import numpy as np

matplotlib.use('Agg')
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import os, sys, inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(currentdir)
sys.path.insert(0, parentdir)
from cami_amber import plots
from cami_amber.utils import labels as utils_labels


def plot_by_genome(data, out_file=None, sort_by='completeness'):
    not_sort_by = list(set(['precision_bp','recall_bp']) - set([sort_by]))[0]  # get the metric not sorted by
    data = sorted(data, key=lambda x: x[sort_by])
    genomes = []
    precision = []
    recall = []
    for genome in data:
        genomes.append(genome['mapped_genome'])
        precision.append(genome['precision_bp'])
        recall.append(genome['recall_bp'])
    sort = {'precision_bp': precision, 'recall_bp': recall}

    fig, ax1 = plt.subplots(figsize=(len(genomes) * 0.15, 5))

    ax1.plot(np.arange(len(genomes)), sort[sort_by], color='black')
    plt.xticks(np.arange(len(genomes)), genomes, rotation='vertical', fontsize="smaller")
    ax1.plot(np.arange(len(genomes)), sort[not_sort_by], '.', color='red')

    # transform y labels to percentages
    vals = ax1.get_yticks()
    ax1.set_yticklabels(['{:3.0f}%'.format(x * 100) for x in vals])

    plt.legend((sort_by.title(), not_sort_by.title()))
    plt.grid(True)
    plt.tight_layout()
    if out_file is None:
        plt.show()
    else:
        plt.savefig(os.path.normpath(out_file + '.png'), dpi=100, format='png', bbox_inches='tight')
        plt.savefig(os.path.normpath(out_file + '.pdf'), dpi=100, format='pdf', bbox_inches='tight')
    plt.close(fig)


def plot_precision_recall_per_bin(pd_bins, output_dir):
    colors_list = plots.create_colors_list()
    df_groups = pd_bins[[utils_labels.TOOL, 'precision_bp', 'recall_bp']].dropna().groupby(utils_labels.TOOL)
    if len(df_groups) > len(colors_list):
        raise RuntimeError("Plot only supports 29 colors")

    fig, axs = plt.subplots(figsize=(6, 5))

    # force axis to be from 0 to 100%
    axs.set_xlim([0.0, 1.0])
    axs.set_ylim([0.0, 1.0])

    # for query_metrics in bin_metrics_per_query:
    for i, (tool, pd_summary) in enumerate(df_groups):
        precision = pd_summary['precision_bp'].tolist()
        recall = pd_summary['recall_bp'].tolist()
        axs.scatter(precision, recall, marker='o', color=colors_list[i], s=[8] * len(precision))

    # turn on grid
    axs.minorticks_on()
    axs.grid(which='major', linestyle='-', linewidth='0.5')
    axs.grid(which='minor', linestyle=':', linewidth='0.5')

    # transform plot_labels to percentages
    vals = axs.get_xticks()
    axs.xaxis.set_major_locator(ticker.FixedLocator(vals))
    axs.set_xticklabels(['{:3.0f}%'.format(x * 100) for x in vals])
    vals = axs.get_yticks()
    axs.yaxis.set_major_locator(ticker.FixedLocator(vals))
    axs.set_yticklabels(['{:3.0f}%'.format(x * 100) for x in vals])

    plt.xlabel('Purity per bin')
    plt.ylabel('Completeness per genome')
    plt.tight_layout()
    fig.savefig(os.path.join(output_dir, 'genome', 'purity_completeness_per_bin.eps'), dpi=100, format='eps', bbox_inches='tight')

    lgd = plt.legend(list(df_groups.groups.keys()), bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0., handlelength=0, frameon=False)
    for handle in lgd.legend_handles:
        handle.set_sizes([100.0])

    fig.savefig(os.path.join(output_dir, 'genome', 'purity_completeness_per_bin.png'), dpi=100, format='png', bbox_extra_artists=(lgd,), bbox_inches='tight')
    fig.savefig(os.path.join(output_dir, 'genome', 'purity_completeness_per_bin.pdf'), dpi=100, format='pdf', bbox_extra_artists=(lgd,), bbox_inches='tight')
    plt.close(fig)



# Copyright 2024 Department of Computational Biology for Infection Research - Helmholtz Centre for Infection Research
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from cami_amber.utils import labels as utils_labels
import pandas as pd


def evaluate_sample(queries_list):
    query1 = queries_list[0]
    gs_data = query1.gold_standard_data
    for query in queries_list:
        query.compute_metrics(gs_data)
    del gs_data


def evaluate_samples_queries(sample_id_to_g_queries_list, sample_id_to_t_queries_list):
    pd_bins = pd.DataFrame()
    df_summary = pd.DataFrame()

    def get_metrics(sample_id_to_queries_list):
        nonlocal pd_bins
        nonlocal df_summary

        for sample_id in sample_id_to_queries_list:
            evaluate_sample(sample_id_to_queries_list[sample_id])

        for sample_id in sample_id_to_queries_list:
            for query in sample_id_to_queries_list[sample_id]:
                if query.eval_success:
                    df_summary = pd.concat([df_summary, query.get_metrics_df().dropna(axis=1, how='all')], ignore_index=True, sort=True)
                    pd_bins = pd.concat([pd_bins, query.precision_df.reset_index()], ignore_index=True, sort=True)

    get_metrics(sample_id_to_g_queries_list)
    get_metrics(sample_id_to_t_queries_list)

    # Gold standard only has unfiltered metrics, so copy values to unfiltered columns
    for col in df_summary.columns:
        if col.endswith(utils_labels.UNFILTERED):
            df_summary.loc[df_summary[utils_labels.TOOL] == utils_labels.GS, col] = \
                df_summary.loc[df_summary[utils_labels.TOOL] == utils_labels.GS, col[:-len(utils_labels.UNFILTERED)]]

    return df_summary, pd_bins



# Copyright 2024 Department of Computational Biology for Infection Research - Helmholtz Centre for Infection Research
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import os
import numpy as np
import pandas as pd

RANKS = ['superkingdom', 'phylum', 'class', 'order', 'family', 'genus', 'species', 'strain']
RANK_TO_INDEX = dict(zip(RANKS, list(range(len(RANKS)))))
INDEX_TO_RANK = dict(zip(list(range(len(RANKS))), RANKS))
RANKS_LOW2HIGH = list(reversed(RANKS))


def load_merged(names_file_path):
    tax_id_to_tax_id = {}
    with open(names_file_path) as read_handler:
        for line in read_handler:
            if len(line.strip()) == 0:
                continue
            line = line.split('|')
            line = list(map(str.strip, line))
            tax_id_to_tax_id[int(line[0])] = int(line[1])
    return tax_id_to_tax_id


def load_names(tax_id_to_rank, names_file_path):
    tax_id_to_name = {}
    with open(names_file_path) as read_handler:
        for line in read_handler:
            if len(line.strip()) == 0:
                continue
            line = line.split('|')
            line = list(map(str.strip, line))
            tax_id = int(line[0])

            if line[3] == "scientific name":
                tax_id_to_name[tax_id] = line[1]
            else:
                continue

            if tax_id_to_rank[tax_id] == "species" or tax_id_to_rank[tax_id] == "strain":
                names = tax_id_to_name[tax_id].split(" ")
                if len(names) > 2:
                    if tax_id_to_rank[tax_id] == "strain":
                        tax_id_to_name[tax_id] = "{} {} strain".format(names[0], names[1])
                    else:
                        tax_id_to_name[tax_id] = "{} {}".format(names[0], names[1])
    tax_id_to_name[""] = ""
    return tax_id_to_name


def check_parent_is_species(tax_id_to_parent, tax_id_to_rank, tax_id):
    if tax_id_to_parent[tax_id] in tax_id_to_parent:
        if tax_id_to_rank[tax_id_to_parent[tax_id]] == "species":
            return True
        elif tax_id_to_parent[tax_id] != 1 and tax_id_to_rank[tax_id_to_parent[tax_id]] not in RANKS:
            return check_parent_is_species(tax_id_to_parent, tax_id_to_rank, tax_id_to_parent[tax_id])
    return False


def load_tax_info(ncbi_nodes_file):
    tax_id_to_parent = {}
    tax_id_to_rank = {}
    with open(ncbi_nodes_file) as read_handler:
        for line in read_handler:
            if len(line.strip()) == 0:
                continue
            line = line.split('|')
            line = list(map(str.strip, line))
            tax_id = int(line[0])
            tax_id_to_parent[tax_id] = int(line[1])
            tax_id_to_rank[tax_id] = line[2]

    for tax_id, rank in tax_id_to_rank.items():
        if tax_id_to_rank[tax_id] == "no rank" and check_parent_is_species(tax_id_to_parent, tax_id_to_rank, tax_id):
            tax_id_to_rank[tax_id] = "strain"

    return tax_id_to_parent, tax_id_to_rank


def get_id_path(tax_id, tax_id_to_parent, tax_id_to_rank, tax_id_to_tax_id):
    if tax_id not in tax_id_to_rank:
        if tax_id_to_tax_id and tax_id in tax_id_to_tax_id:
            tax_id = tax_id_to_tax_id[tax_id]
        else:
            logging.getLogger('amber').warning("Invalid NCBI taxonomic ID: {}".format(tax_id))
            return [np.nan] * 8

    while tax_id_to_rank[tax_id] not in RANKS:
        tax_id = tax_id_to_parent[tax_id]
        if tax_id == 1:
            return [np.nan] * 8

    index = RANK_TO_INDEX[tax_id_to_rank[tax_id]]
    path = [np.nan] * 8
    path[index] = tax_id

    id = tax_id
    try:
        while id in tax_id_to_parent:
            id = tax_id_to_parent[id]
            if id == 1:
                break
            if tax_id_to_rank[id] not in RANKS:
                continue
            index = RANK_TO_INDEX[tax_id_to_rank[id]]
            path[index] = id
            if tax_id_to_rank[id] == 'superkingdom':
                break
    except KeyError:
        logging.getLogger('amber').warning('Could not get rank for taxonomic ID: {}'.format(tax_id))
    return path


def preprocess_ncbi_tax(ncbi_dir):
    def get_path(row):
        return get_id_path(row['TAXID'], tax_id_to_parent, tax_id_to_rank, tax_id_to_tax_id)

    merged = pd.read_csv(os.path.join(ncbi_dir, 'merged.dmp'), sep='|', engine='python', header=None, index_col=False, names=['TAXID'], usecols=[0])
    nodes = pd.read_csv(os.path.join(ncbi_dir, 'nodes.dmp'), sep='|', engine='python', header=None, names=['TAXID'], usecols=[0])
    nodes = pd.concat([nodes, merged], ignore_index=True)

    tax_id_to_parent, tax_id_to_rank = load_tax_info(os.path.join(ncbi_dir, 'nodes.dmp'))
    tax_id_to_tax_id = load_merged(os.path.join(ncbi_dir, 'merged.dmp'))

    nodes = pd.concat([nodes, nodes.apply(lambda row: get_path(row), axis=1, result_type='expand')], axis='columns')
    for i in range(8):
        nodes[i] = nodes[i].astype(pd.UInt32Dtype())
    nodes.rename(columns=INDEX_TO_RANK, inplace=True)

    tax_id_to_name = load_names(tax_id_to_rank, os.path.join(ncbi_dir, 'names.dmp'))
    for k, v in tax_id_to_tax_id.items():
        tax_id_to_name[k] = tax_id_to_name[tax_id_to_tax_id[k]]

    nodes['name'] = nodes['TAXID'].map(tax_id_to_name)

    for k, v in tax_id_to_tax_id.items():
        tax_id_to_rank[k] = tax_id_to_rank[tax_id_to_tax_id[k]]
    nodes['rank'] = nodes.apply(lambda x: tax_id_to_rank[x['TAXID']], axis=1)

    if os.access(ncbi_dir, os.W_OK):
        nodes.to_feather(os.path.join(ncbi_dir, 'nodes.amber.ft'))
    else:
        logging.getLogger('amber').warning('You have no permission to write processed NCBI taxonomy to {}'.format(ncbi_dir))

    return nodes.set_index('TAXID')



#!/usr/bin/env python

# Copyright 2020 Department of Computational Biology for Infection Research - Helmholtz Centre for Infection Research
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import argparse
import os
import gzip
import mimetypes
import sys
from Bio import SeqIO

try:
    import argparse_parents
except ImportError:
    sys.path.append(os.path.dirname(__file__))
    try:
        import argparse_parents
    finally:
        sys.path.remove(os.path.dirname(__file__))


def read_lengths_from_fastx_file(fastx_file):
    """

    @param fastx_file: file path
    @type fastx_file: str
    @rtype: dict[str, int]
    """
    file_type = mimetypes.guess_type(fastx_file)[1]
    if file_type == 'gzip':
        f = gzip.open(fastx_file, "rt")
    elif not file_type:
        f = open(fastx_file, "rt")
    else:
        raise RuntimeError("Unknown type of file: '{}".format(fastx_file))

    length = {}
    if os.path.getsize(fastx_file) == 0:
        return length

    file_format = None
    line = f.readline()
    if line.startswith('@'):
        file_format = "fastq"
    elif line.startswith(">"):
        file_format = "fasta"
    f.seek(0)

    if not file_format:
        raise RuntimeError("Invalid sequence file: '{}".format(fastx_file))

    for seq_record in SeqIO.parse(f, file_format):
        length[seq_record.id] = len(seq_record.seq)

    f.close()

    return length


def add_column(mapping_file, fasta_file):
    lengths = read_lengths_from_fastx_file(fasta_file)
    sequence_id_column_index = 0
    with open(mapping_file, 'r') as read_handler:
        for line in read_handler:
            line = line.rstrip('\n')
            if line.startswith('@@'):
                for index, column_name in enumerate(line[2:].split('\t')):
                    if column_name == 'SEQUENCEID':
                        sequence_id_column_index = index
                print('%s\t_LENGTH' % line)
                continue

            if len(line.strip()) == 0 or line.startswith('#') or line.startswith('@'):
                print(line)
                continue

            row_data = line.split('\t')
            sequence_id = row_data[sequence_id_column_index]
            print('%s\t%d' % (line, lengths[sequence_id]))


def main():
    parser = argparse.ArgumentParser(description="Add length column _LENGTH to gold standard mapping and print mapping on the standard output")
    parser.add_argument("-g", "--gold_standard_file", help=argparse_parents.HELP_GOLD_STANDARD_FILE, required=True)
    parser.add_argument("-f", "--fasta_file", help="FASTA or FASTQ file with sequences of gold standard", required=True)
    args = parser.parse_args()
    add_column(args.gold_standard_file, args.fasta_file)


if __name__ == "__main__":
    main()